package com.io.yutian.elementoriginlib.command;

import com.io.yutian.elementoriginlib.command.argument.Argument;
import com.io.yutian.elementoriginlib.command.argument.ArgumentType;
import com.io.yutian.elementoriginlib.command.interfaces.Command;
import com.io.yutian.elementoriginlib.command.interfaces.Parameter;
import com.io.yutian.elementoriginlib.command.interfaces.SubCommand;
import com.io.yutian.elementoriginlib.exception.command.CommandParseException;
import com.io.yutian.elementoriginlib.logger.Logger;
import org.bukkit.command.CommandSender;

import java.lang.reflect.Method;
import java.util.*;

public class CommandEntity {

    private static final Logger LOGGER = Logger.getLogger(CommandEntity.class);

    private Object instance;

    private String command;
    private String permission;

    private List<CommandEntry> childrens = new ArrayList<>();

    public CommandEntity(Object instance, String command, String permission, List<CommandEntry> childrens) {
        this.instance = instance;
        this.command = command;
        this.permission = permission;
        this.childrens = childrens;
    }

    public Object getInstance() {
        return instance;
    }

    public void addChild(CommandEntry child) {
        childrens.add(child);
    }

    public String getCommand() {
        return command;
    }

    public String getPermission() {
        return permission;
    }

    public List<CommandEntry> getChildrens() {
        return childrens;
    }

    public boolean canInvoke(CommandSender sender) {
        if (childrens.size() == 1 && childrens.get(0).isNodal()) {
            CommandEntry child = childrens.get(0);
            return child.canInvoke(sender);
        }
        return true;
    }

    public boolean hasPermission(CommandSender sender) {
        if (childrens.size() == 1 && childrens.get(0).isNodal()) {
            CommandEntry child = childrens.get(0);
            return child.hasPermission(sender);
        }
        return sender.isOp() || sender.hasPermission(permission);
    }

    public static CommandEntity parseFromClass(Object instance, Class clazz) {
        if (clazz == null) {
            throw new NullPointerException("Class is null");
        }
        if (!clazz.isAnnotationPresent(Command.class)) {
            throw new CommandParseException("类 " + clazz + " 未标注 @Command 注解");
        }

//        Object instance;
//        try {
//            instance = clazz.getConstructor(null).newInstance();
//        } catch (Exception e) {
//            e.printStackTrace();
//            throw new CommandParseException("无法实例化类 "+clazz);
//        }

        Command commandAnnotation = (Command) clazz.getAnnotation(Command.class);
        String command = commandAnnotation.value();
        String permission = commandAnnotation.permission();

        List<CommandEntry> allEntries = new ArrayList<>();
        Map<String, CommandEntry> pathToEntryMap = new HashMap<>();
        List<CommandEntry> rootEntries = new ArrayList<>();

        List<String> allPaths = new ArrayList<>();

        boolean hasNodal = false;

        try {
            for (Method method : clazz.getDeclaredMethods()) {
                if (!method.isAnnotationPresent(SubCommand.class)) {
                    continue;
                }

                CommandEntry commandEntry = parseFromMethod(clazz, method);
                if (commandEntry.isNodal()) {
                    if (hasNodal) {
                        throwParseException(clazz, "命令路径冲突: 命令 '" + command + "' 不能同时出现多个无节点命令(nodal=true)");
                    }
                    hasNodal = true;
                    commandEntry.setName(command);
                }
                if (commandEntry != null) {
                    String path = commandEntry.getFullName();
                    allPaths.add(path);
                    allEntries.add(commandEntry);
                    pathToEntryMap.put(path, commandEntry);
                }
            }

            if (hasNodal && allEntries.size() > 1) {
                throwParseException(clazz, "命令路径冲突: 命令 '" + command + "' 不能同时包含子命令和无节点命令(nodal=true)");
            }

            for (String path : allPaths) {
                String parentPath = getParentPath(path);

                if (!parentPath.isEmpty() && pathToEntryMap.containsKey(parentPath)) {
                    throwParseException(clazz, "命令路径冲突: 子命令 '" + path + "' 已经定义, 不能再定义父命令 '" + parentPath + "'");
                }
            }

            for (CommandEntry entry : allEntries) {
                String fullPath = entry.getFullName();
                int lastDotIndex = fullPath.lastIndexOf('.');

                if (lastDotIndex == -1) {
                    rootEntries.add(entry);
                } else {
                    String parentPath = fullPath.substring(0, lastDotIndex);
                    CommandEntry parentEntry = pathToEntryMap.get(parentPath);

                    if (parentEntry == null) {
                        parentEntry = createParentNodes(pathToEntryMap, parentPath);
                        if (parentPath.indexOf('.') == -1) {
                            rootEntries.add(parentEntry);
                        }
                    }
                    parentEntry.getChildrens().add(entry);
                }
            }

        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }

        return new CommandEntity(instance, command, permission, rootEntries);
    }


    private static String getParentPath(String path) {
        int lastDotIndex = path.lastIndexOf('.');
        if (lastDotIndex == -1) {
            return "";
        }
        return path.substring(0, lastDotIndex);
    }

    private static CommandEntry createParentNodes(Map<String, CommandEntry> pathToEntryMap, String currentPath) {
        if (pathToEntryMap.containsKey(currentPath)) {
            return pathToEntryMap.get(currentPath);
        }

        int lastDotIndex = currentPath.lastIndexOf('.');
        CommandEntry parentEntry;

        if (lastDotIndex == -1) {
            parentEntry = new CommandEntry(currentPath, currentPath);
        } else {
            String grandParentPath = currentPath.substring(0, lastDotIndex);
            CommandEntry grandParentEntry = createParentNodes(pathToEntryMap, grandParentPath);
            parentEntry = new CommandEntry(currentPath, currentPath.substring(lastDotIndex + 1));
            grandParentEntry.getChildrens().add(parentEntry);
        }
        pathToEntryMap.put(currentPath, parentEntry);
        return parentEntry;
    }


    private static CommandEntry parseFromMethod(Class clazz, Method method) {
        SubCommand subCommand = method.getAnnotation(SubCommand.class);
        String path = subCommand.value();
        String permission = subCommand.permission();
        boolean nodal = subCommand.nodal();
        String[] senderRequireArray = subCommand.senderRequire();
        List<SenderRequire> senderRequireList = new ArrayList<>();
        for (String senderRequire : senderRequireArray) {
            SenderRequire senderRequire1 = SenderRequires.get(senderRequire);
            if (senderRequire1 == null) {
                throwParseException(clazz, "使用者权限 " + senderRequire + " 不存在");
            }
            senderRequireList.add(SenderRequires.get(senderRequire));
        }

        String subName = path;

        if (path.contains(".")) {
            int lastDotIndex = path.lastIndexOf('.');
            if (lastDotIndex == -1 || lastDotIndex > path.length() - 1) {
                throwParseException(clazz, "命令名 '" + path + "' 无效");
            }
            subName = path.substring(path.lastIndexOf('.') + 1);
        }
        List<Argument> arguments = new ArrayList<>();
        int allArgumentsCount = (int) Arrays.stream(method.getParameters())
                .filter(parameter -> parameter.isAnnotationPresent(Parameter.class))
                .count();
        int index = 0;
        for (java.lang.reflect.Parameter parameter : method.getParameters()) {
            if (!parameter.isAnnotationPresent(Parameter.class)) {
                continue;
            }
            Parameter commandArgument = parameter.getAnnotation(Parameter.class);
            String name = commandArgument.name();
            String argumentName = name.isEmpty() ? parameter.getName() : name;
            boolean required = commandArgument.required();
            if (!required && index < allArgumentsCount - 1) {
                throwParseException(clazz, "可选参数仅能在最后一个位置中使用");
            }
            String defaultValue = commandArgument.defaultValue();
            String suggestionType = commandArgument.suggestionType();
            Class<?> type = parameter.getType();
            ArgumentType argumentType = ArgumentType.get(parameter.getType());
            if (argumentType == null) {
                throwParseException(clazz, "参数类型 " + parameter.getType().getName() + " 不存在");
            }
            Argument argument = new Argument(argumentName, argumentType);
            if (!suggestionType.isEmpty()) {
                Suggest suggest = Suggests.getSuggest(suggestionType);
                if (suggest == null) {
                    throwParseException(clazz, "建议类型 " + parameter.getType().getName() + " 不存在");
                } else {
                    argument.suggest(suggest);
                }
            }
            if (!required) {
                argument.optional(defaultValue);
            }
            arguments.add(argument);
            index++;
        }
        return new CommandEntry(path, subName, permission, nodal, senderRequireList, arguments, method);
    }

    private static void throwParseException(Class clazz, String message) {
        LOGGER.warn("解析命令 "+clazz+" 时出现错误");
        throw new CommandParseException(message);
    }

    @Override
    public String toString() {
        return "CommandEntity{" +
                "command='" + command + '\'' +
                ", permission='" + permission + '\'' +
                ", childrens=" + childrens +
                '}';
    }

}
